APP_NAME = 'drivnal'
APP_NAME_FORMATED = 'Drivnal'
CONF_FILENAME = '%s.conf' % APP_NAME

CLOSED = 'closed'
SAVED = 'saved'
UNSAVED = 'unsaved'

FAILED = 'failed'
ABORTED = 'aborted'
ABORTING = 'aborting'
PENDING = 'pending'
COMPLETE = 'complete'

CREATE_SNAPSHOT = 'create_snapshot'
RESTORE_OBJECT = 'restore_object'
MOVE_VOLUME = 'move_volume'

VOLUMES_UPDATED = 'volumes_updated'
SNAPSHOTS_UPDATED = 'snapshots_updated'
TASKS_UPDATED = 'tasks_updated'

SNAPSHOT_DIR = 'snapshots'
LOG_DIR = 'logs'

DEFAULT_MIN_FREE_SPACE = 0.1
DEFAULT_MAX_PRUNE = 0.8
DEFAULT_MAX_RETRY = 3
DEFAULT_DB_PATH = '/var/lib/drivnal/drivnal.db'
DEFAULT_WWW_PATH = '/usr/share/drivnal/www'

DEFAULT_ROOT_EXCLUDES = [
    'dev/*',
    'proc/*',
    'sys/*',
    'tmp/*',
    'run/*',
    'mnt/*',
    'media/*',
    'lost+found',
]

DIR_MIME_TYPE = 'inode/directory'

MIN_UNITS = ['m', 'min', 'minute']
HOUR_UNITS = ['h', 'hr', 'hour']
DAY_UNITS = ['d', 'day']
WEEKDAY_UNITS = {
    'monday': 0,
    'mon': 0,
    'mo': 0,
    'm': 0,
    'tuesday': 1,
    'tues': 1,
    'tue': 1,
    'tu': 1,
    't': 1,
    'wednesday': 2,
    'wed': 2,
    'we': 2,
    'w': 2,
    'thursday': 3,
    'thur': 3,
    'thu': 3,
    'th': 3,
    'friday': 4,
    'fri': 4,
    'fr': 4,
    'f': 4,
    'saturday': 5,
    'sat': 5,
    'sa': 5,
    's': 5,
    'sunday': 6,
    'sun': 6,
    'su': 6,
}

VOLUME_NOT_FOUND = 'volume_not_found'
VOLUME_NOT_FOUND_MSG = 'Volume not found'

SNAPSHOT_NOT_FOUND = 'snapshot_not_found'
SNAPSHOT_NOT_FOUND_MSG = 'Snapshot not found'

PATH_NOT_FOUND = 'path_not_found'



class SnapshotError(Exception):
    pass

class SnapshotPermissionError(SnapshotError):
    pass

class SnapshotAlreadyRunning(SnapshotError):
    pass

class SnapshotPathExists(SnapshotError):
    pass

class SnapshotSyncError(SnapshotError):
    pass

class SnapshotSpaceError(SnapshotError):
    pass

class SnapshotMoveError(SnapshotError):
    pass

class RestoreError(Exception):
    pass

class RestoreSyncError(RestoreError):
    pass



from constants import *
import os
import urllib
import mimetypes
import logging

class Object:
    def __init__(self, path):
        self.name = os.path.basename(path)
        self.path = path

        if os.path.isdir(self.path):
            self.type = DIR_MIME_TYPE
            self.size = None
            self.time = None
        else:
            try:
                stat = os.stat(self.path)
                self.size = stat.st_size
                self.time = stat.st_mtime
            except OSError:
                self.size = None
                self.time = None

            mime = mimetypes.MimeTypes()
            self.type = mime.guess_type(urllib.pathname2url(self.path))[0]

    @staticmethod
    def get_objects(path):
        objects = []

        if path:
            for name in os.listdir(path):
                object_path = os.path.join(path, name)
                objects.append(Object(object_path))

        return objects



from constants import *
from config import Config
from volume import Volume
from drivnal import server
import os
import copy
import logging

logger = logging.getLogger(APP_NAME)

class Client:
    def __init__(self):
        pass

    def __getattr__(self, name):
        if name in ['volumes']:
            self.load()

        if name not in self.__dict__:
            raise AttributeError('Object instance has no attribute %r' % name)
        return self.__dict__[name]

    def get_volume(self, id):
        for volume in self.volumes:
            if volume.id == id:
                return volume

    def create_volume(self, path):
        logger.debug('Creating volume.')

        if not os.path.isdir(path):
            os.mkdir(path)
        volume = Volume(self, path)

        self.volumes.append(volume)
        logger.debug('Adding volume path to config. %r' % {
            'volume_id': volume.id,
        })
        self.commit()

        return volume

    def remove_volume(self, volume):
        logger.debug('Removing volume. %r' % {
            'volume_id': volume.id,
        })

        self.volumes.remove(volume)
        self.commit()

    def load(self):
        logger.debug('Loading client database.')
        self.volumes = []

        for i, volume_path in enumerate(
                server.app_db.get('system', 'volumes')):

            try:
                volume_path = os.path.normpath(volume_path)
            except AttributeError:
                logger.error('Failed to normalize volume path. %r' % {
                    'volume_num': i,
                })
            try:
                volume = Volume(self, volume_path)
                self.volumes.append(volume)
            except IOError:
                logger.debug('Failed to load volume. %r' % {
                    'volume_num': i,
                })

    def commit(self):
        logger.debug('Writing client database.')

        server.app_db.remove('system', 'volumes')
        for volume in self.volumes:
            server.app_db.set('system', 'volumes', volume.path, None)



from constants import *
from drivnal import server
import logging
import time
import uuid

logger = logging.getLogger(APP_NAME)

_STR_DATABASE_VARIABLES = ['volume_id', 'type']
_INT_DATABASE_VARIABLES = ['time']

class Event:
    def __init__(self, id=None, volume_id=None, type=None):
        if id is None:
            self.id = uuid.uuid1().hex
            self.type = type
            self.time = int(time.time() * 1000)
            self.volume_id = None
        else:
            self.id = id

        if volume_id is not None:
            self.volume_id = volume_id

    def __setattr__(self, name, value):
        if name in _STR_DATABASE_VARIABLES:
            server.app_db.set('events', self.id, name, value)
        elif name in _INT_DATABASE_VARIABLES:
            server.app_db.set('events', self.id, name, str(value))
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        if name in _STR_DATABASE_VARIABLES:
            return server.app_db.get('events', self.id, name)
        elif name in _INT_DATABASE_VARIABLES:
            value = server.app_db.get('events', self.id, name)
            if value:
                return int(value)
            return None
        elif name not in self.__dict__:
            raise AttributeError('Object instance has no attribute %r' % name)
        return self.__dict__[name]

    @staticmethod
    def _validate(data):
        if 'time' not in data or not data['time']:
            return False

        try:
            data_time = int(data['time'])
        except ValueError:
            return False

        if 'type' not in data or not data['type']:
            return False

        if 'volume_id' not in data or not data['volume_id']:
            return False

        return True

    @staticmethod
    def get_events(volume, last_time):
        events = []
        events_dict = {}
        events_time = []

        logger.debug('Getting events for volume. %r' % {
            'volume_id': volume.id,
        })

        events_query = server.app_db.get('events')
        for event_id in events_query:
            event = events_query[event_id]

            # Remove broken events
            if not Event._validate(event):
                logger.debug('Removing broken event from database. %r' % {
                    'volume_id': volume.id,
                    'event_id': event_id,
                })
                server.app_db.remove('events', event_id)
                continue

            event['time'] = int(event['time'])

            if event['time'] <= last_time:
                continue

            if event['volume_id'] != volume.id:
                continue

            # Prevent events with the same time from breaking sorted list,
            # event that is sent to client will always have original time
            while True:
                if event['time'] not in events_time:
                    break
                event['time'] += 1

            events_dict[event['time']] = Event(id=event_id)
            events_time.append(event['time'])

        for event_time in sorted(events_time):
            events.append(events_dict[event_time])

        return events



from constants import *
from event import Event
from drivnal import server
import threading
import logging
import time
import uuid

logger = logging.getLogger(APP_NAME)
task_threads = {}

_STR_DATABASE_VARIABLES = ['volume_id', 'type', 'state']
_INT_DATABASE_VARIABLES = ['time']

class Task:
    def __init__(self, id=None, volume=None):
        if id is None:
            self.id = uuid.uuid1().hex
            self.type = None
            self.state = None
            self.time = None
            self.volume_id = None
            task_threads[self.id] = None
        else:
            self.id = id

        if volume is None:
            self.volume = None
        else:
            self.volume_id = volume.id
            self.volume = volume

        self.update()

    def __setattr__(self, name, value):
        if name == 'thread':
            task_threads[self.id] = value
        elif name in _STR_DATABASE_VARIABLES:
            server.app_db.set('tasks', self.id, name, value)
        elif name in _INT_DATABASE_VARIABLES:
            server.app_db.set('tasks', self.id, name, str(value))
        else:
            self.__dict__[name] = value

        if name == 'state' and value:
            Event(volume_id=self.volume_id, type=TASKS_UPDATED)

    def __getattr__(self, name):
        if name == 'thread':
            if self.id not in task_threads:
                return None
            return task_threads[self.id]
        elif name in _STR_DATABASE_VARIABLES:
            return server.app_db.get('tasks', self.id, name)
        elif name in _INT_DATABASE_VARIABLES:
            value = server.app_db.get('tasks', self.id, name)
            if value:
                return int(value)
            return None
        elif name not in self.__dict__:
            raise AttributeError('Object instance has no attribute %r' % name)
        return self.__dict__[name]

    def abort(self):
        logger.debug('Aborting task. %r' % {
            'volume_id': self.volume_id,
            'task_id': self.id,
        })
        self.state = ABORTING
        Event(volume_id=self.volume_id, type=TASKS_UPDATED)

    def aborted(self):
        logger.debug('Task aborted. %r' % {
            'volume_id': self.volume_id,
            'task_id': self.id,
        })
        self.state = ABORTED
        self.thread = None
        Event(volume_id=self.volume_id, type=TASKS_UPDATED)

    def remove(self):
        logger.debug('Removing task. %r' % {
            'volume_id': self.volume_id,
            'task_id': self.id,
        })
        server.app_db.remove('tasks', self.id)
        Event(volume_id=self.volume_id, type=TASKS_UPDATED)

    def run(self, *args, **kwargs):
        raise NotImplementedError('Derived class must override run method')

    def _run(self, *args, **kwargs):
        try:
            # 0 second tasks causes error
            time.sleep(1)

            self.run(*args, **kwargs)
            if self.state == PENDING:
                self.state = COMPLETE
            elif self.state == ABORTING:
                self.state = ABORTED
            self.thread = None
        except:
            logger.exception('Task failed. %r' % {
                'volume_id': self.volume_id,
                'task_id': self.id,
            })
            self.state = FAILED
            self.thread = None
        finally:
            if self.state == PENDING:
                self.state = FAILED
                self.thread = None

    def start(self, *args, **kwargs):
        self.volume_id = self.volume_id
        self.time = int(time.time())
        self.state = PENDING
        self.thread = threading.Thread(
            target=self._run, args=args, kwargs=kwargs)
        self.thread.start()

    def update(self):
        if self.state not in [PENDING, ABORTING]:
            return

        logger.debug('Updating task state. %r' % {
            'volume_id': self.volume_id,
            'task_id': self.id,
        })

        if not self.thread or not self.thread.is_alive():
            logger.warning('Task failed, thread ended unexpectable. %r' % {
                'volume_id': self.volume_id,
                'task_id': self.id,
            })
            self.state = FAILED

    @staticmethod
    def _validate(data):
        if 'time' not in data or not data['time']:
            return False

        try:
            data_time = int(data['time'])
        except ValueError:
            return False

        if 'type' not in data or not data['type']:
            return False

        if 'volume_id' not in data or not data['volume_id']:
            return False

        if 'state' not in data or not data['state']:
            return False

        return True

    @staticmethod
    def _get_tasks(volume, type, states=[]):
        tasks = []
        tasks_dict = {}
        tasks_time = []

        logger.debug('Getting tasks for volume. %r' % {
            'volume_id': volume.id,
        })

        tasks_query = server.app_db.get('tasks')
        for task_id in tasks_query:
            task = tasks_query[task_id]

            # Remove broken events
            if not Task._validate(task):
                logger.debug('Removing broken task from database. %r' % {
                    'volume_id': volume.id,
                    'task_id': task_id,
                })
                server.app_db.remove('tasks', task_id)
                continue

            task['time'] = int(task['time'])

            if type and task['type'] != type:
                continue
            if task['volume_id'] != volume.id:
                continue
            if states and task['state'] not in states:
                continue
            if not states and not task['state']:
                continue

            # Task update occurs here recheck state
            task = Task(id=task_id)
            if states and task.state not in states:
                continue
            tasks_dict[task.time] = task
            tasks_time.append(task.time)

        for task_time in sorted(tasks_time):
            tasks.append(tasks_dict[task_time])

        return tasks

    @staticmethod
    def get_tasks(volume, type=None):
        return Task._get_tasks(volume, type)

    @staticmethod
    def get_failed_tasks(volume, type=None):
        return Task._get_tasks(volume, type, [FAILED])

    @staticmethod
    def get_aborted_tasks(volume, type=None):
        return Task._get_tasks(volume, type, [ABORTED])

    @staticmethod
    def get_pending_tasks(volume, type=None):
        return Task._get_tasks(volume, type, [PENDING, ABORTING])

    @staticmethod
    def get_completed_tasks(volume, type=None):
        return Task._get_tasks(volume, type, [COMPLETE])



from constants import *
from exceptions import *
from task import Task
import os
import time
import logging
import subprocess

logger = logging.getLogger(APP_NAME)

class RestoreObject(Task):
    def __init__(self, *kargs, **kwargs):
        Task.__init__(self, *kargs, **kwargs)
        self.type = RESTORE_OBJECT

    def _abort_process(self, process):
        for i in xrange(10):
            process.terminate()
            logger.debug('Terminating restore process. %r' % {
                'volume_id': self.volume_id,
                'pid': process.pid,
            })
            time.sleep(1)
            if process.poll() is not None:
                break

        if process.poll() is None:
            for i in xrange(30):
                process.kill()
                logger.debug('Killing restore process. %r' % {
                    'volume_id': self.volume_id,
                    'pid': process.pid,
                })
                time.sleep(0.5)
                if process.poll() is not None:
                    break

        if process.poll() is None:
            logger.error('Failed to abort restore process. %r' % {
                'volume_id': self.volume_id,
                'pid': process.pid,
            })

        self.aborted()

    def run(self, objects, destination_path):
        logger.info('Restoring objects. %r' % {
            'volume_id': self.volume_id,
        })

        log_path = os.path.join(self.volume.path,
            LOG_DIR, 'restore_%s.log' % int(time.time()))

        for obj in objects:
            args = ['rsync', '--archive', '--hard-links', '--acls',
                '--quiet', '--xattrs',  '--progress', '--super',
                '--log-file-format=%o \"%f\" %l', '--log-file=%s' % log_path,
                obj.path, destination_path]

            process = subprocess.Popen(args)
            return_code = None

            while True:
                poll = process.poll()
                if poll is not None:
                    return_code = poll
                    break
                if self.state == ABORTING:
                    self._abort_process(process)
                    return
                time.sleep(0.1)

            if return_code != 0:
                raise RestoreSyncError('Object restore failed, command ' + \
                    'returned non-zero exit status. %r' % {
                        'volume_id': self.volume_id,
                        'snapshot_id': snapshot_id,
                    })



from constants import *
from exceptions import *
from task import Task
from event import Event
import os
import time
import shlex
import logging
import subprocess

logger = logging.getLogger(APP_NAME)

class CreateSnapshot(Task):
    def __init__(self, *kargs, **kwargs):
        Task.__init__(self, *kargs, **kwargs)
        self.type = CREATE_SNAPSHOT

    def _get_free_space(self, path):
        path_stat = os.statvfs(self.volume.path)
        return 1 - (1 - float(path_stat.f_bavail) /
            float(path_stat.f_blocks))

    def _check_no_space_error(self, path, log_path):
        logger.debug('Checking for no space error on %r.' % self.volume.name)

        with open(log_path) as log:
            # Get last kilobyte of file
            log.seek(0, os.SEEK_END)
            file_size = log.tell()
            log.seek(max(file_size - 1024, 0))
            lines = log.readlines()

        no_space_error = False
        for line in lines:
            if ': No space left on device (28)' in line and \
                    'write failed on' in line:
                no_space_error = True

                # Get destination file path from log then replace snapshot
                # path with source path to get the source path of file
                line = line[line.find('write failed on') + 15:line.find(
                    ': No space left on device (28)')]
                file_path = shlex.split(line)[0]
                file_path = file_path.replace(path + '.temp',
                    self.volume.source_path)

                if os.path.isfile(file_path):
                    try:
                        file_stat = os.stat(file_path)
                        return (no_space_error, file_stat.st_size)
                    except OSError:
                        logger.warning('Failed to get size of file, ' + \
                            'that caused no space error. %r' % {
                                'volume_id': self.volume_id,
                                'snapshot_id': self.snapshot_id,
                            })
                break

        return (no_space_error, None)

    def remove_snapshot(self):
        self.volume.load_snapshots()
        snapshot = self.volume.get_snapshot(self.snapshot_id)
        if snapshot:
            self.volume.remove_snapshot(snapshot)
        else:
            logger.debug('Failed to remove snapshot, snapshot id does ' + \
                'not exists.')

    def _prune_snapshot(self, snapshot):
        logger.info('Volume low on space, auto removing snapshot. %r' % {
            'volume_id': self.volume_id,
            'snapshot_id': self.snapshot_id,
            'removing_snapshot_id': snapshot.id,
        })
        self.volume.remove_snapshot(snapshot)

    def prune_snapshots(self, min_size=None):
        snapshot_count = self.volume.get_snapshot_count()
        min_free_space = self.volume.min_free_space or DEFAULT_MIN_FREE_SPACE
        max_prune = self.volume.max_prune or DEFAULT_MAX_PRUNE
        max_prune_count = (int(self.orig_snapshot_count * max_prune) -
            (self.orig_snapshot_count - snapshot_count))
        orig_free_space = self.volume.get_space_free()

        logger.debug('Pruning snapshots. %r' % {
            'volume_id': self.volume_id,
            'snapshot_id': self.snapshot_id,
        })

        # If a min space needed is set and percent of volume is greater then
        # min_free_space use the percent needed + 0.01 for min free space
        percent_needed = None
        if min_size:
            path_stat = os.statvfs(self.volume.path)
            percent_needed = min_size / float(
                path_stat.f_bsize * path_stat.f_blocks) + 0.01

            if percent_needed > min_free_space:
                min_free_space = percent_needed

        for snapshot in self.volume.get_failed_snapshots():
            if self.volume.get_space_free() >= min_free_space:
                return False
            self._prune_snapshot(snapshot)

        # Remove up to the max prune count of volumes until required free
        # space has been reached
        for i in xrange(max_prune_count):
            if self.volume.get_space_free() >= min_free_space:
                return False
            self._prune_snapshot(self.volume.snapshots[0])

        if orig_free_space != self.volume.get_space_free():
            return True
        return False

    def _abort_process(self, process):
        for i in xrange(10):
            process.terminate()
            logger.debug('Terminating snapshot process. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'pid': process.pid,
            })
            time.sleep(1)
            if process.poll() is not None:
                break

        if process.poll() is None:
            for i in xrange(30):
                process.kill()
                logger.debug('Killing snapshot process. %r' % {
                    'volume_id': self.volume_id,
                    'snapshot_id': self.snapshot_id,
                    'pid': process.pid,
                })
                time.sleep(0.5)
                if process.poll() is not None:
                    break

        if process.poll() is None:
            logger.error('Failed to abort snapshot process. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'pid': process.pid,
            })

        logger.warning('Snapshot aborted, removing aborted snapshot. %r' % {
            'volume_id': self.volume_id,
            'snapshot_id': self.snapshot_id,
        })

        self.remove_snapshot()
        self.aborted()
        Event(volume_id=self.volume_id, type=VOLUMES_UPDATED)

    def run(self):
        self.snapshot_id = int(time.time())
        self.orig_snapshot_count = self.volume.get_snapshot_count()
        bandwidth_limit = self.volume.bandwidth_limit
        excludes = self.volume.excludes
        source_path = self.volume.source_path
        try:
            destination_path = os.path.join(self.volume.path,
                SNAPSHOT_DIR, str(self.snapshot_id))
        except AttributeError:
            logger.error('Failed to join snapshot destination path. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'task_id': self.id,
            })
        destination_path_temp = destination_path + '.temp'
        destination_path_failed = destination_path + '.failed'
        log_path = os.path.join(self.volume.path,
            LOG_DIR, 'snapshot_%s.log' % self.snapshot_id)
        max_retry = self.volume.max_retry or DEFAULT_MAX_RETRY
        last_snapshot = self.volume.get_last_snapshot()

        logger.info('Creating snapshot. %r' % {
            'volume_id': self.volume_id,
            'snapshot_id': self.snapshot_id,
            'task_id': self.id,
        })

        Event(volume_id=self.volume_id, type=VOLUMES_UPDATED)

        snapshots_path = os.path.join(self.volume.path, SNAPSHOT_DIR)
        logs_path = os.path.join(self.volume.path, LOG_DIR)

        if not os.path.isdir(snapshots_path):
            logger.debug('Creating volume snapshots directory. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'task_id': self.id,
            })
            os.mkdir(snapshots_path)

        if not os.path.isdir(logs_path):
            logger.debug('Creating volume logs directory. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'task_id': self.id,
            })
            os.mkdir(logs_path)

        args = ['rsync', '--archive', '--delete', '--hard-links',
            '--acls', '--quiet', '--xattrs',  '--progress', '--super',
            '--log-file-format=%o \"%f\" %l', '--log-file=%s' % log_path]

        if bandwidth_limit:
            args.append('--bwlimit=%s' % bandwidth_limit)

        if last_snapshot:
            args.append('--link-dest=%s' % last_snapshot.path)

        rsync_source_path = source_path
        if rsync_source_path[-1] != os.sep:
            rsync_source_path += os.sep

        if rsync_source_path == os.sep:
            logger.debug('Auto excluding default paths for root backup. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'task_id': self.id,
            })
            for exclude in DEFAULT_ROOT_EXCLUDES:
                if exclude in excludes:
                    continue
                excludes.append(exclude)

        # If volume is a subdirectory of source path exclude path
        if os.path.commonprefix(
                [rsync_source_path, self.volume.path]) == rsync_source_path:
            logger.debug('Auto excluding volume path from snapshot. %r' % {
                'volume_id': self.volume_id,
                'snapshot_id': self.snapshot_id,
                'task_id': self.id,
            })
            excludes.append(self.volume.path.replace(rsync_source_path, '', 1))

        if excludes:
            for exclude in excludes:
                args.append('--exclude=%s' % exclude)

        args.append(rsync_source_path)
        args.append(destination_path_temp)

        for path in [destination_path, destination_path_temp,
                destination_path_failed]:
            if os.path.isdir(path):
                raise SnapshotPathExists('Snapshot failed, snapshot ' + \
                    'path already exists. %r' % {
                        'volume_id': self.volume_id,
                        'snapshot_id': self.snapshot_id,
                        'task_id': self.id,
                    })

        self.prune_snapshots()

        for i in xrange(max_retry):
            process = subprocess.Popen(args)
            return_code = None

            while True:
                poll = process.poll()
                if poll is not None:
                    return_code = poll
                    break
                if self.state == ABORTING:
                    self._abort_process(process)
                    return
                time.sleep(0.1)

            if return_code != 0:
                logger.debug('Command returned non-zero exit status. %r' % {
                    'volume_id': self.volume_id,
                    'snapshot_id': self.snapshot_id,
                    'task_id': self.id,
                    'cmd_args': args,
                    'return_code': return_code,
                })

                # Check if error was from no space
                no_space_error, last_file_size = self._check_no_space_error(
                    destination_path, log_path)
                if no_space_error and i <= max_retry:
                    # If prune was able to free up space try again
                    if self.prune_snapshots():
                        logger.warning('Snapshot did not have required ' + \
                            'free space, retrying snapshot. %r' % {
                                'volume_id': self.volume_id,
                                'snapshot_id': self.snapshot_id,
                                'task_id': self.id,
                            })
                        continue

                try:
                    os.rename(destination_path_temp, destination_path_failed)
                except OSError:
                    logger.error('Unable to rename failed snapshot. %r' % {
                        'volume_id': self.volume_id,
                        'snapshot_id': self.snapshot_id,
                        'task_id': self.id,
                    })

                if no_space_error:
                    raise SnapshotSpaceError('Snapshot failed, unable ' + \
                        'to free up required space. %r' % {
                            'volume_id': self.volume_id,
                            'snapshot_id': self.snapshot_id,
                            'task_id': self.id,
                            'cmd_args': args,
                            'return_code': return_code,
                        })
                else:
                    raise SnapshotSyncError('Snapshot failed, command ' + \
                        'returned non-zero exit status. %r' % {
                            'volume_id': self.volume_id,
                            'snapshot_id': self.snapshot_id,
                            'task_id': self.id,
                            'cmd_args': args,
                            'return_code': return_code,
                        })
            break

        try:
            os.rename(destination_path_temp, destination_path)
        except OSError:
            raise SnapshotMoveError('Snapshot failed, unable to rename ' + \
                'snapshot temp directory. %r' % {
                    'volume_id': self.volume_id,
                    'snapshot_id': self.snapshot_id,
                    'task_id': self.id,
                })

        Event(volume_id=self.volume_id, type=VOLUMES_UPDATED)
        Event(volume_id=self.volume_id, type=SNAPSHOTS_UPDATED)



import os
import flask
import cherrypy.wsgiserver
import logging
from constants import *
from database import Database
from werkzeug import SharedDataMiddleware

logger = None

class Server:
    def __init__(self):
        self.config = None
        self.conf_path = None
        self.app = None
        self.app_db = None

    def _setup_app(self):
        self.app = flask.Flask(APP_NAME)

        global logger
        logger = self.app.logger

        from config import Config
        self.config = Config(self.conf_path)
        self.config.read()

        if self.config.log_debug == 'true':
            self.log_level = logging.DEBUG
        else:
            self.log_level = logging.INFO

        if self.config.log_path:
            self.log_handler = logging.FileHandler(self.config.log_path)
        else:
            self.log_handler = logging.StreamHandler()

        logger.setLevel(self.log_level)
        self.log_handler.setLevel(self.log_level)

        self.log_handler.setFormatter(logging.Formatter(
            '[%(asctime)s][%(levelname)s][%(module)s][%(lineno)d] ' +
            '%(message)s'))

        logger.addHandler(self.log_handler)

        self.app_db = Database(self.config.db_path or DEFAULT_DB_PATH)

        import handlers

    def _setup_static(self):
        www_path = self.config.www_path or DEFAULT_WWW_PATH

        self.app.wsgi_app = SharedDataMiddleware(self.app.wsgi_app, {
            '/': os.path.normpath(www_path),
        }, cache=False)

        @self.app.route('/', methods=['GET'])
        def index_get():
            with open(os.path.join(www_path, 'index.html'), 'r') as fd:
                return fd.read()

    def _start_debug(self):
        scheduler = None
        if self.config.scheduler != 'false':
            from scheduler import Scheduler
            scheduler = Scheduler()
            scheduler.start()

        # App.run server uses werkzeug logger
        werkzeug_logger = logging.getLogger('werkzeug')
        werkzeug_logger.setLevel(self.log_level)
        werkzeug_logger.addHandler(self.log_handler)

        try:
            self.app.run(host=self.config.bind_addr,
                port=int(self.config.port), threaded=True)
        finally:
            logger.info('Stopping server...')
            if scheduler:
                self.app_db.sync()
                scheduler.stop()
            self.app_db.sync()

    def _start(self):
        scheduler = None
        if self.config.scheduler != 'false':
            from scheduler import Scheduler
            scheduler = Scheduler()
            scheduler.start()

        server = cherrypy.wsgiserver.CherryPyWSGIServer(
            (self.config.bind_addr, int(self.config.port)), self.app)
        try:
            server.start()
        except (KeyboardInterrupt, SystemExit), exc:
            logger.info('Stopping server...')
            self.app_db.sync()
            if scheduler:
                scheduler.stop()
            server.stop()
            self.app_db.sync()

    def start(self):
        self._setup_app()
        self._setup_static()

        logger.info('Starting server...')

        if self.config.debug == 'true':
            self._start_debug()
        else:
            self._start()



from constants import *
from client import Client
import time
import logging
import threading

class Scheduler:
    def __init__(self):
        self.interrupt = None
        self.thread = None

    def _run(self):
        last_check = time.localtime().tm_min

        while not self.interrupt:
            localtime = time.localtime()

            if localtime.tm_min == last_check:
                time.sleep(0.5)
                continue

            last_check = localtime.tm_min
            self.check_schedule(localtime)

    def _create_snapshot(self, volume):
        volume.create_snapshot()

    def _check_volume_schedule(self, localtime, volume):
        schedule = volume.schedule

        try:
            time_num = int(''.join([x for x in schedule if x.isdigit()]))
        except ValueError:
            time_num = None
        time_unit = ''.join([x for x in schedule if x.isalpha()]).lower()

        if time_unit in WEEKDAY_UNITS:
            if (localtime.tm_hour == 4 and
                    localtime.tm_min == 0 and
                    localtime.tm_wday == WEEKDAY_UNITS[time_unit]):
                self._create_snapshot(volume)
        else:
            time_unit = time_unit.replace('s', '')

        if time_unit in MIN_UNITS:
            if (int(time.mktime(localtime)) / 60) % time_num == 0:
                self._create_snapshot(volume)
        elif time_unit in HOUR_UNITS:
            if ((int(time.mktime(localtime)) / 60 / 60) % time_num == 0
                    and localtime.tm_min == 0):
                self._create_snapshot(volume)
        elif time_unit in DAY_UNITS:
            if (localtime.tm_hour == 4 and
                    localtime.tm_min == 0 and
                    ((localtime.tm_yday - 1) % time_num) == 0):
                self._create_snapshot(volume)

    def check_schedule(self, localtime):
        client = Client()

        for volume in client.volumes:
            self._check_volume_schedule(localtime, volume)

    def start(self):
        self.interrupt = None
        self.thread = threading.Thread(
            target=self._run)
        self.thread.start()

    def stop(self):
        self.interrupt = True
        self.thread.join()



from constants import *
from exceptions import *
from config import Config
from origin import Origin
from snapshot import Snapshot
from task import Task
from event import Event
from create_snapshot import CreateSnapshot
from restore_object import RestoreObject
from move_volume import MoveVolume
import os
import time
import uuid
import shutil
import logging

logger = logging.getLogger(APP_NAME)

class Volume:
    def __init__(self, client, path):
        try:
            config_path = os.path.join(path, CONF_FILENAME)
        except AttributeError:
            logger.error('Failed to join volume config path. %r' % {
                'volume_id': self.id,
            })
            raise

        self.config = Config(config_path)
        self.client = client

        self.orig_path = path
        self.path = path

        if not self.id:
            self.id = uuid.uuid1().hex
            logger.debug('Generated and setting volume uuid. %r' % {
                'volume_id': self.id,
            })
            self.commit()

    def __getattr__(self, name):
        if name in ['id', 'name', 'source_path', 'excludes', 'schedule', \
                'min_free_space', 'bandwidth_limit', 'origin']:
            self.load()
        if name in ['snapshots']:
            self.load_snapshots()

        if name not in self.__dict__:
            raise AttributeError('Object instance has no attribute %r' % name)
        return self.__dict__[name]

    def get_snapshot(self, id):
        for snapshot in self.snapshots:
            if snapshot.id == id:
                return snapshot

    def get_failed_snapshots(self):
        snapshots = []
        for snapshot in self.snapshots:
            if snapshot.state == FAILED:
                snapshots.append(snapshot)
        return snapshots

    def get_snapshots(self):
        snapshots = []
        for snapshot in self.snapshots:
            if snapshot.state == COMPLETE:
                snapshots.append(snapshot)
        return snapshots

    def get_snapshot_count(self):
        count = 0
        for snapshot in self.snapshots:
            if snapshot.state == COMPLETE:
                count += 1
        return count

    def get_last_snapshot(self):
        for snapshot in reversed(self.snapshots):
            if snapshot.state == COMPLETE:
                return snapshot

    def create_snapshot(self):
        logger.debug('Creating snapshot. %r' % {
            'volume_id': self.id,
        })

        if not os.access(self.path, os.W_OK):
            raise SnapshotPermissionError('Snapshot failed, write access ' + \
                'denied to volume storage directory. %r' % {
                    'volume_id': self.id,
                })

        if self.snapshot_pending():
            raise SnapshotAlreadyRunning('Snapshot failed, ' + \
                'snapshot already running. %r' % {
                    'volume_id': self.id,
                })

        logger.debug('Starting snapshot task. %r' % {
            'volume_id': self.id,
        })

        task = CreateSnapshot(volume=self)
        task.start()

        return task

    def remove_snapshot(self, snapshot, keep_log=False):
        logger.info('Removing snapshot. %r' % {
            'volume_id': self.id,
            'snapshot_id': snapshot.id,
        })

        self.snapshots.index(snapshot)

        if not keep_log:
            logger.debug('Removing snapshot log file. %r' % {
                'volume_id': self.id,
                'snapshot_id': snapshot.id,
            })

            if os.path.isfile(snapshot.log_path):
                os.remove(snapshot.log_path)

        logger.debug('Removing snapshot directory. %r' % {
            'volume_id': self.id,
            'snapshot_id': snapshot.id,
        })

        if os.path.isdir(snapshot.path):
            shutil.rmtree(snapshot.path)

        self.snapshots.remove(snapshot)
        Event(volume_id=self.id, type=SNAPSHOTS_UPDATED)

    def snapshot_pending(self):
        tasks = Task.get_pending_tasks(self, CREATE_SNAPSHOT)
        if tasks:
            return True
        return False

    def get_space_free(self):
        stat = os.statvfs(self.path)
        return float(stat.f_bavail) / stat.f_blocks

    def get_space_used(self):
        return 1 - self.get_space_free()

    def get_tasks(self):
        return Task.get_tasks(self)

    def get_events(self, last_time):
        return Event.get_events(self, last_time)

    def restore_object(self, objects, destination_path):
        logger.debug('Starting restore object task. %r' % {
            'volume_id': self.id,
        })

        task = RestoreObject(volume=self)
        task.start(objects, destination_path)

        return task

    def load(self):
        logger.debug('Reading volume config.')

        try:
            self.config.read()
        except IOError:
            pass

        self.id = self.config.id
        self.name = self.config.name
        self.source_path = self.config.source_path
        if self.source_path:
            try:
                self.source_path = os.path.normpath(self.source_path)
            except AttributeError:
                logger.error('Failed to normalize volume source path. %r' % {
                    'volume_id': self.id,
                })
        self.excludes = self.config.excludes or []
        for i, exclude in enumerate(self.excludes):
            try:
                self.excludes[i] = os.path.normpath(exclude)
            except AttributeError:
                logger.error('Failed to normalize volume exclude path. %r' % {
                    'volume_id': self.id,
                    'exclude_num': i,
                })
        self.schedule = self.config.schedule

        min_free_space = None
        if self.config.min_free_space:
            try:
                min_free_space = float(self.config.min_free_space)
            except ValueError:
                logger.warning(
                    'Config option min_free_space is invalid. %r' % {
                        'volume_id': self.id,
                    })
        self.min_free_space = min_free_space

        bandwidth_limit = None
        if self.config.bandwidth_limit:
            try:
                bandwidth_limit = int(self.config.bandwidth_limit)
            except ValueError:
                logger.warning(
                    'Config option bandwidth_limit is invalid. %r' % {
                        'volume_id': self.id,
                    })
        self.bandwidth_limit = bandwidth_limit

        max_prune = None
        if self.config.max_prune:
            try:
                max_prune = float(self.config.max_prune)
            except ValueError:
                logger.warning('Config option max_prune is invalid. %r' % {
                    'volume_id': self.id,
                })
        self.max_prune = max_prune

        max_retry = None
        if self.config.max_retry:
            try:
                max_retry = int(self.config.max_retry)
            except ValueError:
                logger.warning('Config option max_retry is invalid. %r' % {
                    'volume_id': self.id,
                })
        self.max_retry = max_retry

        self.origin = Origin(self)

    def load_snapshots(self):
        logger.debug('Loading snapshots. %r' % {
            'volume_id': self.id,
        })

        snapshots_path = os.path.join(self.path, SNAPSHOT_DIR)
        snapshot_names = []
        self.snapshots = []

        if not os.path.isdir(snapshots_path):
            return

        for name in os.listdir(snapshots_path):
            if not os.path.isdir(os.path.join(snapshots_path, name)) or \
                    not name.replace('.temp', '').replace(
                        '.failed', '').isdigit() or len(name) < 6:
                continue
            snapshot_names.append(name)

        for name in sorted(snapshot_names):
            self.snapshots.append(Snapshot(self, name))

    def _move_volume(self):
        logger.debug('Starting move volume task. %r' % {
            'volume_id': self.id,
        })

        task = MoveVolume(volume=self)
        task.start()

        return task

    def commit(self):
        retry_config_commit = False

        logger.debug('Writing volume config. %r' % {
            'volume_id': self.id,
        })

        self.config.id = self.id
        self.config.name = self.name
        self.config.source_path = self.source_path
        self.config.excludes = self.excludes
        self.config.schedule = self.schedule
        self.config.min_free_space = self.min_free_space
        self.config.bandwidth_limit = self.bandwidth_limit
        self.config.max_prune = self.max_prune
        self.config.max_retry = self.max_retry
        try:
            self.config.write()
        except:
            # If volume has already been moved write will need
            # to be done after move
            if self.orig_path != self.path:
                logger.debug('Failed to commit config to original ' + \
                    'volume path in a volume move, retying to new path ' +
                    'after volume is moved. %r' % {
                        'volume_id': self.id,
                    })
                retry_config_commit = True
            else:
                raise

        if self.orig_path != self.path:
            self._move_volume()

            if retry_config_commit:
                self.config.set_path(self.path)
                self.config.write()



from constants import *
import threading
import logging
import anydbm
import copy

logger = logging.getLogger(APP_NAME)

class Database():
    def __init__(self, db_path):
        logger.debug('Opening database...')
        self._db = anydbm.open(db_path, 'c')
        self._sync_lock = False

    def __del__(self):
        self._db.sync()

    def _get_prefix(self, prefix):
        prefix_len = len(prefix)
        items = {}

        # Handle RuntimeError if database changes size while iterating
        for i in xrange(256):
            try:
                for item in self._db:
                    if item.startswith(prefix):
                        items[item[prefix_len:]] = self._db[item]
                break
            except RuntimeError:
                if i == 255:
                    raise

        return items

    def _async_sync(self):
        if self._sync_lock:
            return
        threading.Thread(target=self.sync).start()

    def sync(self):
        if self._sync_lock:
            return
        self._sync_lock = True

        logger.debug('Syncing database...')
        self._db.sync()

        self._sync_lock = False

    def get(self, column_family, row=None, column=None):
        if not row and column:
            raise TypeError('Must specify a row for column')

        if not row:
            items = {}
            prefix_items = self._get_prefix('%s-' % column_family)

            for item in prefix_items:
                item_split = item.split('-', 1)
                if item_split[0] not in items:
                    items[item_split[0]] = {}
                items[item_split[0]][item_split[1]] = prefix_items[item]

            return items

        if not column:
            return self._get_prefix('%s-%s-' % (column_family, row))

        item_name = '%s-%s-%s' % (column_family, row, column)
        if item_name in self._db:
            return self._db[item_name]

    def set(self, column_family, row, column, value):
        self._db['%s-%s-%s' % (column_family, row, column)] = value
        self._async_sync()

    def remove(self, column_family, row=None, column=None):
        if not row and column:
            raise TypeError('Must specify a row for column')

        item_name = '%s-' % column_family
        if row:
            item_name += '%s-' % row

        if not column:
            items = []

            # Handle RuntimeError if database changes size while iterating
            for i in xrange(256):
                try:
                    for item in self._db:
                        if item.startswith(item_name):
                            items.append(item)
                    break
                except RuntimeError:
                    if i == 255:
                        raise

            for item in items:
                if item in self._db:
                    del self._db[item]

        else:
            item_name += column
            if item_name in self._db:
                del self._db[item_name]
